
import unittest
from koltrast.chunks import Chunk, parse_chunk

class TestChunk(unittest.TestCase):
    def test_is_valid(self):
        self.assertTrue(Chunk.is_valid("day"))
        self.assertTrue(Chunk.is_valid("hour"))
        self.assertTrue(Chunk.is_valid("week"))
        self.assertTrue(Chunk.is_valid("month"))
        self.assertTrue(Chunk.is_valid("year"))
        self.assertFalse(Chunk.is_valid("invalid"))

    def test_parse_chunk_valid(self):
        self.assertEqual(parse_chunk("day"), Chunk.DAY)
        self.assertEqual(parse_chunk("days"), Chunk.DAY)
        self.assertEqual(parse_chunk("hour"), Chunk.HOUR)
        self.assertEqual(parse_chunk("hours"), Chunk.HOUR)
        self.assertEqual(parse_chunk("week"), Chunk.WEEK)
        self.assertEqual(parse_chunk("weeks"), Chunk.WEEK)
        self.assertEqual(parse_chunk("month"), Chunk.MONTH)
        self.assertEqual(parse_chunk("months"), Chunk.MONTH)
        self.assertEqual(parse_chunk("year"), Chunk.YEAR)
        self.assertEqual(parse_chunk("years"), Chunk.YEAR)

    def test_parse_chunk_invalid(self):
        with self.assertRaises(ValueError):
            parse_chunk("invalid")


if __name__ == '__main__':
    unittest.main()




""" Interval """
from dataclasses import dataclass
from pendulum import DateTime
from typing import List
from pendulum import parse
from koltrast.chunks import Chunk, parse_chunk, add_chunk, is_half_chunk


@dataclass
class Interval:
    """Interval

    since: interval since when? (inclusive: =>)
    until: interval until when? (not inclusive: <)
    """

    since: DateTime
    until: DateTime

    def __init__(self, since: DateTime | str, until: DateTime | str):

        if isinstance(since, str):
            since=parse(since)

        if isinstance(until, str):
            until=parse(until)

        if since >= until:
            raise ValueError("Since must be < until")

        self.since = since
        self.until = until


def _split_interval(
    interval: Interval,
    chunk: Chunk
) -> List[Interval]:
    """ Split an interval into smaller intervals

    Parameters:
        interval: an unfixed amount of time
        chunk: a fixed amount of time

    Returns:
        List of intervals
    """

    sub_intervals = []

    start_here = interval.since
    end_here = interval.until

    while start_here < end_here:

        sub_interval_start = start_here
        sub_interval_end = add_chunk(_datetime=start_here, chunk=chunk)

        if is_half_chunk(start=start_here, end=interval.until, chunk=chunk):
            sub_interval_end = interval.until

        sub_intervals.append(
            Interval(since=sub_interval_start, until=sub_interval_end)
        )

        start_here = sub_interval_end

    return sub_intervals


def generate_intervals(since: str, until: str, chunk: str) -> List[Interval]:
    """ Create generate a list of intervals from an interval
    Parameters:
        since: The lower bound of an interval. Inclusive (>=)
        until: The upper bound of an interval. Not inclusive (<)
        chunk: How big or small should the intervals be.
            Accepted: HOUR, DAY, WEEK, MONTH, YEAR, FULL

    Returns: List of intervals
    """

    return _split_interval(
        interval=Interval(since=since, until=until),
        chunk=parse_chunk(string=chunk)
    )



from enum import Enum
from pendulum import DateTime


class Chunk(str, Enum):
    """ A chunk is a fixed amount of time, unless it is Chunk.FULL, then it is variable """
    DAY = 'day'
    HOUR = 'hour'
    WEEK = 'week'
    MONTH = 'month'
    YEAR = 'year'

    @classmethod
    def is_valid(cls, value: str) -> bool:
        """"""
        return value in {item.value for item in cls}

def parse_chunk(string: str) -> Chunk:
    """
    Parse string to a valid chunk
    """
    string = string.lower()
    if string.endswith("s"):
        string = string[:-1]

    if not Chunk.is_valid(value=string):
        allowed_chunks = ([inc.value for inc in Chunk])
        raise ValueError(f"Unsupported chunk: {string}. Choose one of {allowed_chunks}")

    return Chunk(value=string)


def add_chunk(_datetime: DateTime, chunk: Chunk) -> DateTime:
    if chunk == Chunk.HOUR:
        return _datetime.add(hours=1)
    if chunk == Chunk.DAY:
        return _datetime.add(days=1)
    if chunk == Chunk.WEEK:
        return _datetime.add(weeks=1)
    if chunk == Chunk.MONTH:
        return _datetime.add(months=1)
    if chunk == Chunk.YEAR:
        return _datetime.add(years=1)


def is_half_chunk(start: DateTime, end: DateTime, chunk: Chunk) -> bool:
    return add_chunk(_datetime=start, chunk=chunk) > end



import unittest
from pendulum import datetime
from koltrast.interval import Interval, _split_interval, generate_intervals
from koltrast.chunks import Chunk
from pendulum.exceptions import ParserError

class TestInterval(unittest.TestCase):
    def test_since_is_earlier_than_until(self):
        since = datetime(2023, 1, 1)
        until = datetime(2023, 1, 2)
        interval = Interval(since, until)
        self.assertEqual(interval.since, since)
        self.assertEqual(interval.until, until)

        since = datetime(2023, 1, 2)
        until = datetime(2023, 1, 1)
        with self.assertRaises(ValueError):
            Interval(since, until)

    def test_since_is_earlier_than_until_with_strings(self):
        since = "2023-01-01"
        until = "2023-01-02"
        interval = Interval(since, until)
        self.assertEqual(interval.since, datetime(2023, 1, 1, tz='UTC'))
        self.assertEqual(interval.until, datetime(2023, 1, 2, tz='UTC'))

        since = "2023-01-02"
        until = "2023-01-01"
        with self.assertRaises(ValueError):
            Interval(since, until)

    def test_since_equals_until(self):
        since = "2023-01-01"
        until = "2023-01-01"
        with self.assertRaises(ValueError):
            Interval(since, until)

    def test_invalid_since_string(self):
        since = "invalid-date"
        until = "2023-01-01"
        with self.assertRaises(ParserError):
            Interval(since, until)

    def test_invalid_until_string(self):
        since = "2023-01-01"
        until = "invalid-date"
        with self.assertRaises(ParserError):
            Interval(since, until)

    def test_invalid_both_strings(self):
        since = "invalid-date"
        until = "another-invalid-date"
        with self.assertRaises(ParserError):
            Interval(since, until)

    def test_valid_datetime_objects(self):
        since = datetime(2023, 1, 1)
        until = datetime(2023, 1, 2)
        interval = Interval(since, until)
        self.assertEqual(interval.since, since)
        self.assertEqual(interval.until, until)

    def test_valid_date_strings(self):
        since = "2023-01-01"
        until = "2023-01-02"
        interval = Interval(since, until)
        self.assertEqual(interval.since, datetime(2023, 1, 1, tz='UTC'))
        self.assertEqual(interval.until, datetime(2023, 1, 2, tz='UTC'))


class TestSplitInterval(unittest.TestCase):
    def test_split_one_and_half_weeks(self):
        since = datetime(2023, 1, 1)
        until = datetime(2023, 1, 11)
        chunk = Chunk.WEEK
        interval = Interval(since=since, until=until)
        result = _split_interval(interval, chunk)

        expected_intervals= [
            Interval(datetime(2023, 1, 1),datetime(2023, 1, 8)),
            Interval(datetime(2023, 1, 8),datetime(2023, 1, 11))
            ]

        self.assertEqual(result, expected_intervals)


    def test_split_half_year(self):
        since = datetime(2023, 1, 1)
        until = datetime(2023, 7, 1)
        chunk = Chunk.YEAR

        result = generate_intervals(since, until, chunk)

        expected_intervals = [
            Interval(datetime(2023, 1, 1), datetime(2023, 7, 1))
        ]

        self.assertEqual(result, expected_intervals)

    def test_split_one_month(self):
        since = datetime(2023, 1, 1)
        until = datetime(2023, 2, 1)
        chunk = Chunk.MONTH
        result = generate_intervals(since, until, chunk)

        expected_interval = [Interval(datetime(2023, 1, 1), datetime(2023, 2, 1))]

        self.assertEqual(result, expected_interval)

    def test_inclusive(self):
        since = datetime(2023, 1, 1)
        until = datetime(2023, 1, 31)
        chunk = Chunk.DAY

        result = len(generate_intervals(since, until, chunk))

        expected_len = 30

        self.assertEqual(result, expected_len)

if __name__ == '__main__':
    unittest.main()
